/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2005-2008 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */

/*!
 *  \file   toppers/c_expr.hpp
 *  \brief  C̎̍\͂Ɋւ錾`
 *
 *  ̃t@CŒ`NX
 *  \code
 *  struct c_expr_parser_base< Derived >;
 *  struct c_expr_parser;
 *  struct c_const_expr_parser;
 *  \endcode
 */
#ifndef TOPPERS_C_EXPR_HPP_
#define TOPPERS_C_EXPR_HPP_

#include "toppers/c_parser.hpp"
#include <boost/cstdint.hpp>

namespace toppers
{

  /*!
   *  \struct c_expr_parser_base c_expr.hpp "toppers/c_expr.hpp"
   *  \brief  C̎\͂邽߂̊NX
   *
   *  ۂɎgpۂ́A c_expr_parser_base  c_expr_parser_base::definition
   *  NXhKv܂B c_expr_parser_base::definition ̔hNXł́A
   *  start o֐`āA\̕Kvȃ[o悤ɂĂB
   *  邱ƂŁA萔̕@Aꎟ̕@Ƃ悤ɁAC̕@̃Tu
   *  ZbgeՂɍoƂł܂B
   *
   *  \code
   *  // ꎟo
   *  struct c_primary_expression : c_expr_parser_base< c_primary_expression >
   *  {
   *    template < class Scanner >
   *    struct definition : c_expr_parser_base< c_primary_expression >::definition
   *    {
   *      rule_t const& start() const { return primary_expression; }
   *    };
   *  };
   *  \endcode
   */
  template < class Derived >
    struct c_expr_parser_base : boost::spirit::grammar< Derived >
  {
  public:
    /*!
     *  \struct definition c_expr.hpp "toppers/c_expr.hpp"
     *  \brief  @`
     */
    template < class Scanner >
      struct definition
    {
      typedef boost::spirit::rule< Scanner > rule_t;
      rule_t  primary_expression,   
              expression,
              constant_expression,
              conditional_expression,
              assignment_expression,
              assignment_operator,
              postfix_expression,
              unary_expression,
              unary_operator,
              cast_expression,
              multiplicative_expression,
              additive_expression,
              shift_expression,
              relational_expression,
              equality_expression,
              AND_expression,
              exclusive_OR_expression,
              inclusive_OR_expression,
              logical_AND_expression,
              logical_OR_expression,
              string_literal,
              constant,
              floating_constant,
              decimal_floating_constant,
              hexadecimal_floating_constant,
              integer_constant,
              character_constant,
              declaration_specifiers,
              type_name,
              specifier_qualifier_list,
              storage_class_specifier,
              type_specifier,
              type_qualifier,
              declarator,
              direct_declarator,
              struct_or_union_specifier,
              struct_declaration,
              struct_declarator,
              enum_specifier,
              enumerator,
              abstract_declarator,
              pointer,
              parameter_type_list,
              parameter_list,
              parameter_declaration,
              direct_abstract_declarator;

      c_ident_parser_t identifier;
      c_strlit_parser_t c_strlit_p;
      c_chlit_parser_t c_chlit_p;

      /*!
       *  \brief  RXgN^
       *  \param  self  @NXi c_expr_parser_base< Derived > NX̌pjւ̎Q
       */
      definition( Derived const& self )
        : identifier( c_ident_parser( self.ucn_, self.c_plus_plus_ ) ),
          c_strlit_p( c_strlit_parser( self.codeset_ ) ),
          c_chlit_p( c_chlit_parser( self.codeset_ ) )
      {
        using namespace boost::spirit;
        static functor_parser< detail::c_integer_constant_parse_functor< boost::uintmax_t > > const c_int_const_p;
        static functor_parser< detail::c_integer_suffix_parse_functor > const c_int_suffix_p;

        primary_expression =  // eΉ
            identifier
          | constant
          | string_literal
          | ( '(' >> expression >> ')' );
        expression =
            assignment_expression % ',';
        constant_expression = 
            conditional_expression;
        conditional_expression =
            logical_OR_expression >> *( '\?' >> expression >> ':' >> logical_OR_expression );
        assignment_expression =
           *( unary_expression >> assignment_operator ) >> conditional_expression;
        assignment_operator =
            ch_p( '=' ) | "*=" | "/=" | "%=" | "+=" | "?=" | "<<=" | ">>=" | "&=" | "^=" | "|=";
        postfix_expression =
            primary_expression >>
          *(
                ( '[' >> expression >> ']' )
              | ( '(' >> list_p( assignment_expression, ',' ) >> ')' )
              | ( '.' >> identifier )
              | ( "->" >> identifier )
              | "++"
              | "--"
            );
        unary_expression =
           *( str_p( "++" ) || "--" ) >>
            (
                ( "sizeof" >> unary_expression )
              | ( str_p( "sizeof" ) >> '(' >> type_name >> ')' )
              | postfix_expression
              | ( unary_operator >> cast_expression )
            );
        unary_operator =
            chset<>( "&*~!+-" );
        cast_expression =
            *( '(' >> type_name >> ')' ) >> unary_expression
          | +( '(' >> ( type_name | identifier ) >> ')' );  // \͂Ɏss΍
        multiplicative_expression =
            cast_expression >>
           *(
                ( '*' >> cast_expression )
              | ( '/' >> cast_expression )
              | ( '%' >> cast_expression )
            );
        additive_expression =
            multiplicative_expression >>
           *(
                ( '+' >> multiplicative_expression )
              | ( '-' >> multiplicative_expression )
            );
        shift_expression =
            additive_expression >>
           *(
                ( "<<" >> additive_expression )
              | ( ">>" >> additive_expression )
            );
        relational_expression =
            shift_expression >>
           *(
                ( '<' >> shift_expression )
              | ( '>' >> shift_expression )
              | ( "<=" >> shift_expression )
              | ( ">=" >> shift_expression )
            );
        equality_expression =
            relational_expression >>
           *(
                ( "==" >> relational_expression )
              | ( "!=" >> relational_expression )
            );
        AND_expression =
            equality_expression >> *( '&' >> equality_expression );
        exclusive_OR_expression =
            AND_expression >> *( '^' >> AND_expression );
        inclusive_OR_expression =
            exclusive_OR_expression >> *( '|' >> exclusive_OR_expression );
        logical_AND_expression =
            inclusive_OR_expression >> *( "&&" >> inclusive_OR_expression );
        logical_OR_expression =
            logical_AND_expression >> *( "||" >> logical_AND_expression );
        string_literal =
            c_strlit_p
          | lexeme_d[ 'L' >> c_strlit_p ];
        constant =
            floating_constant
          | integer_constant
          | identifier  // 񋓒萔
          | character_constant;
        floating_constant =
            decimal_floating_constant
          | hexadecimal_floating_constant;
        decimal_floating_constant =
            lexeme_d
            [
              as_lower_d
              [
                ( ( *digit_p >> '.' >> +digit_p ) | ( +digit_p >> '.' ) ) >>
                'e' >> !chset<>( "+-" ) >> +digit_p >>
                !chset<>( "fl" )
              ]
            ];
        hexadecimal_floating_constant =
            lexeme_d
            [
              as_lower_d
              [
                "0x" >>
                ( ( *xdigit_p >> '.' >> +xdigit_p ) | ( +xdigit_p >> '.' ) ) >>
                'p' >> !chset<>( "+-" ) >> +digit_p >>
                !chset<>( "fl" )
              ]
            ];
        integer_constant =
            lexeme_d[ c_int_const_p >> !c_int_suffix_p ];
        character_constant =
            c_chlit_p
          | lexeme_d[ 'L' >> c_chlit_p ];
        declaration_specifiers =
           +( storage_class_specifier | type_specifier | type_qualifier );
        type_name =
            specifier_qualifier_list >> !abstract_declarator;
        specifier_qualifier_list =
           +( type_specifier | type_qualifier );
        storage_class_specifier =
            str_p( "auto" )
          | "register"
          | "static"
          | "extern"
          | "typedef";
        type_specifier =
            str_p( "void" ) | "char" | "short" | "int" | "long" | "float" | "double"
          | "signed" | "unsigned"
          | identifier
          | struct_or_union_specifier
          | enum_specifier;
        type_qualifier =
            str_p( "const" ) | "volatile" | "restrict";
        declarator =
            !pointer >> direct_declarator;
        direct_declarator = 
            ( identifier | ( '(' >> declarator >> ')' ) )
            >>
          *(
                ( '[' >> !constant_expression >> ']' )
              | ( '(' >> parameter_type_list >> ')' )
              | ( '(' >> !( identifier % ',' ) >> ')' )
            );
        struct_or_union_specifier =
            lexeme_d[ ( str_p( "struct" ) | "union" ) >> +space_p >> identifier ]
          | ( lexeme_d[ ( str_p( "struct" ) | "union" ) >> +space_p >> !identifier ] >> '{' >> +struct_declaration >> '}' );
        struct_declaration =
            specifier_qualifier_list >> !list_p( struct_declarator, ',' ) >> ';';
            // lisp_p( struct_declarator, ',' )ȗ\ƂĂ̂́A
            // struct_declarator  identifier  specifier_qualifier_list 
            // typedef ƊԈႤƂ邽
        struct_declarator =
            ( !declarator >> ':' >> constant_expression ) // rbgtB[h
          | declarator;
        enum_specifier =
            ( lexeme_d[ "enum" >> +space_p >> !identifier ] >> '{' >> list_p( enumerator, ',', ',' ) >> '}' )   // C99ł͖̃J}Ă悢
          | lexeme_d[ "enum" >> +space_p >> identifier ];
        enumerator =
            identifier >> !( '=' >> constant_expression );
        abstract_declarator =
            ( !pointer >> direct_abstract_declarator )
          | pointer;
        pointer =
           +( '*' >> *type_qualifier );
        parameter_type_list =
            parameter_list >> !( ch_p( ',' ) >> "..." );  // ό
        parameter_list =
            parameter_declaration % ',';
        parameter_declaration = 
            ( declaration_specifiers >> declarator )
          | ( declaration_specifiers >> !abstract_declarator );
        direct_abstract_declarator =
            (
             !( '(' >> abstract_declarator >> ')' ) >>
             +(
                  ( '[' >> !constant_expression >> ']' )
                | ( '(' >> !parameter_type_list >> ')' ) 
              )
            )
          | ( '(' >> abstract_declarator >> ')' );
      }
    };
    bool ucn_;
    codeset_t codeset_;
    bool c_plus_plus_;

    /*!
     *  \brief  RXgN^
     *  \param  ucn         ەɑΉꍇ true w肷
     *  \param  codeset     R[h
     *  \param  c_plus_plus C++ ɑΉꍇ true w肷
     */
    explicit c_expr_parser_base( bool ucn = false, codeset_t codeset = ascii, bool c_plus_plus = false )
      : ucn_( ucn ), codeset_( codeset ), c_plus_plus_( c_plus_plus )
    {
    }
  };

  /*!
   *  \class  c_expr_parser c_expr.hpp "toppers/c_expr.hpp" 
   *  \brief  C̎̍\̓NX
   */
  struct c_expr_parser : c_expr_parser_base< c_expr_parser >
  {
    typedef c_expr_parser_base< c_expr_parser > base_t;

    /*!
     *  \struct definition c_expr.hpp "toppers/c_expr.hpp"
     *  \brief  @`
     */
    template < class Scanner >
    struct definition : base_t::definition< Scanner >
    {
      typedef typename base_t::definition< Scanner >::rule_t rule_t;

      definition( c_expr_parser const& self ) : base_t::definition< Scanner >( self ) {}
      rule_t const& start() const { return base_t::definition< Scanner >::expression; };
    };

    /*!
     *  \brief  RXgN^
     *  \param  ucn     ەɑΉꍇ true w肷
     *  \param  codeset R[h
     */
    explicit c_expr_parser( bool ucn = false, codeset_t codeset = ascii )
      : c_expr_parser_base< c_expr_parser >( ucn, codeset )
    {
    }
  };

  /*!
   *  \class  c_const_expr_parser c_expr.hpp "toppers/c_expr.hpp"
   *  \brief  C̒萔̍\̓NX
   */
  struct c_const_expr_parser : c_expr_parser_base< c_const_expr_parser >
  {
    typedef c_expr_parser_base< c_const_expr_parser > base_t;

    /*!
     *  \struct definition c_expr.hpp "toppers/c_expr.hpp"
     *  \brief  @`
     */
    template < class Scanner >
    struct definition : base_t::definition< Scanner >
    {
      typedef typename base_t::definition< Scanner >::rule_t rule_t;

      definition( c_const_expr_parser const& self ) : base_t::definition< Scanner >( self ) {}
      rule_t const& start() const { return base_t::definition< Scanner >::constant_expression; };
    };

    /*!
     *  \brief  RXgN^
     *  \param  ucn     ەɑΉꍇ true w肷
     *  \param  codeset R[h
     */
    explicit c_const_expr_parser( bool ucn = false, codeset_t codeset = ascii )
      : c_expr_parser_base< c_const_expr_parser >( ucn, codeset )
    {
    }
  };

}

#endif  // ! TOPPERS_C_EXPR_HPP_
